# -*- coding: utf-8 -*-
"""
Authors: Laust Færch
Module: Collect/PROBAV

Description:
This module preprocesses Landsat(7+8) data downloaded externaly

"""
from pywapor.collect.product.Landsat import C2L2SP

__all__ = ["C2L2SP"]

__version__ = '0.1'
