# -*- coding: utf-8 -*-
"""
Authors: Laust Færch
Module: Collect/PROBAV

Description:
This module preprocesses Landsat(7+8) data downloaded externaly

"""
from pywapor.collect.product import Landsat, CHIRPS, ERA5, GEOS5, GLOBCOVER, MERRA2, MODIS, PROBAV, SRTM, STATICS, SENTINEL2, SENTINEL3, VIIRSL1, COPERNICUS

__all__ = ["Landsat", "CHIRPS", "ERA5", "GEOS5", "GLOBCOVER", "MERRA2", "MODIS", "PROBAV", "SRTM", "STATICS", "SENTINEL2", "SENTINEL3", "VIIRSL1", "COPERNICUS"]

__version__ = '0.1'
